''' Script to check pylandau properties.
'''
import unittest

import numpy as np
import pylandau

from hypothesis import given, settings
import hypothesis.strategies as st
from scipy.integrate import quad as integrate
from scipy.optimize import fmin

from tests import constraints


class TestProperties(unittest.TestCase):

    def test_landau_pdf_integral(self):
        ''' Check that Landau pdf integral is 1 '''
        result, _ = integrate(pylandau.get_landau_pdf,
                              0, 10000, args=(10, 1))
        self.assertAlmostEqual(result, 1, delta=1e-3)

    def test_langau_pdf_integral(self):
        ''' Check that Langau pdf integral is 1 '''
        result, _ = integrate(pylandau.get_langau_pdf,
                              -10000, 10000, args=(10, 1, 3))
        self.assertAlmostEqual(result, 1, delta=1e-3)

    def test_gauss_pdf_integral(self):
        ''' Check that Gauss pdf integral is 1 '''
        result, _ = integrate(pylandau.get_gauss_pdf,
                              0, 10000, args=(10, 3))
        self.assertAlmostEqual(result, 1, delta=1e-3)

    @given(st.floats(constraints.LANDAU_PDF_MIN_MU,
                     constraints.LANDAU_PDF_MAX_MU,
                     allow_nan=False,
                     allow_infinity=False))
    def test_landau_pdf_mu(self, mu):
        ''' Check that Landau pdf mu position '''
        # For the given parameters (mu=mu, eta=1) MPV is expected
        # at x = mu-0.22278
        # See also: https://root.cern.ch/root/html524/TMath.html#TMath:Landau

        # x0 > 1 otherwise scipy.optimize.fmin terminates too early
        if abs(mu) > 1.:
            x0 = mu
        else:
            x0 = 1. * np.sign(mu)

        mpv = fmin(lambda x: -pylandau.get_landau_pdf(x, mu=mu, eta=1.),
                   x0=x0, full_output=True, disp=False,
                   xtol=0.000001, ftol=0.000001)[0][0]
        self.assertAlmostEqual(mpv, mu - 0.22278, delta=1e9)

    @given(st.tuples(
        # mpv
        st.floats(constraints.LANDAU_MIN_MPV,
                  constraints.LANDAU_MAX_MPV,
                  allow_nan=False,
                  allow_infinity=False),
        # eta
        st.floats(constraints.LANDAU_MIN_ETA,
                  constraints.LANDAU_MAX_ETA,
                  allow_nan=False,
                  allow_infinity=False),
        # A
        st.floats(constraints.LANDAU_MIN_A,
                  constraints.LANDAU_MAX_A,
                  allow_nan=False,
                  allow_infinity=False),)
           )
    @settings(deadline=None)  # Test is known to take some time
    def test_landau_fallback(self, pars):
        ''' Check if langau is landau for sigma = 0 '''
        (mpv, eta, A) = pars
        x = np.linspace(mpv - 5 * eta, mpv + 5 * eta, 1000)
        y_1 = pylandau.landau(x, mpv=mpv, eta=eta, A=A)
        y_2 = pylandau.langau(x, mpv=mpv, eta=eta, sigma=0., A=A)
        self.assertTrue(np.all(y_1 == y_2))


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestProperties)
    unittest.TextTestRunner(verbosity=2).run(suite)



''' Script to check pylandau exceptions.
'''
import unittest

import numpy as np
import pylandau


class TestExceptions(unittest.TestCase):

    def test_negative_amplitude(self):
        ''' Check exception for negative amplitude '''
        x = np.linspace(0, 100)
        with self.assertRaises(ValueError):
            pylandau.landau(x, A=-1)

        x = np.linspace(0, 100)
        with self.assertRaises(ValueError):
            pylandau.langau(x, A=-1)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestExceptions)
    unittest.TextTestRunner(verbosity=2).run(suite)



''' Simple fit without fit error estimation.
For correct fit errors check the advanced example.
Bound should be defined. Eta has to be > 1.
'''

from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
import numpy as np
import pylandau

# Create fake data with possion error
mpv, eta, sigma, A = 30, 5, 4, 1000
x = np.arange(0, 100, 0.5)
y = pylandau.langau(x, mpv, eta, sigma, A)
yerr = np.random.normal(np.zeros_like(x), np.sqrt(y))
yerr[y < 1] = 1
y += yerr

# Fit with constrains
coeff, pcov = curve_fit(pylandau.langau, x, y,
                        sigma=yerr,
                        absolute_sigma=True,
                        p0=(mpv, eta, sigma, A),
                        bounds=(1, 10000))

# Plot
plt.errorbar(x, y, np.sqrt(pylandau.langau(x, *coeff)), fmt=".")
plt.plot(x, pylandau.langau(x, *coeff), "-")
plt.show()



''' Check pylandau with random inputs
'''
import unittest

import numpy as np
import pylandau

from hypothesis import given, assume
import hypothesis.strategies as st

from tests import constraints


class TestFuzzing(unittest.TestCase):

    @given(st.floats(constraints.LANDAU_MIN_MPV,
                     constraints.LANDAU_MAX_MPV,
                     allow_nan=False,
                     allow_infinity=False))
    def test_landau_mpv(self, mpv):
        ''' Check Landau MPV position '''
        x = np.linspace(mpv - 10, mpv + 10, 1000)
        y = pylandau.landau(x, mpv=mpv, eta=1., A=1.)
        delta = x[1] - x[0]
        self.assertAlmostEqual(x[np.argmax(y)], mpv, delta=delta)

    @given(st.floats(constraints.LANGAU_MIN_MPV,
                     constraints.LANGAU_MAX_MPV,
                     allow_nan=False,
                     allow_infinity=False))
    def test_langau_mpv(self, mpv):
        ''' Check Langau MPV position '''
        x = np.linspace(mpv - 10, mpv + 10, 1000)
        y = pylandau.langau(x, mpv=mpv, eta=1., sigma=1., A=1.)
        delta = x[1] - x[0]
        self.assertAlmostEqual(x[np.argmax(y)], mpv, delta=delta)

    @given(st.floats(constraints.LANDAU_MIN_A,
                     constraints.LANDAU_MAX_A,
                     allow_nan=False,
                     allow_infinity=False))
    def test_landau_A(self, A):
        ''' Check Landau amplitude '''
        mpv = 1.
        x = np.linspace(mpv - 10, mpv + 10, 1000)
        y = pylandau.landau(x, mpv=mpv, eta=1., A=A)
        self.assertAlmostEqual(y.max(), A, delta=1e-4 * A)

    @given(st.floats(constraints.LANGAU_MIN_A,
                     constraints.LANGAU_MAX_A,
                     allow_nan=False,
                     allow_infinity=False))
    def test_langau_A(self, A):
        ''' Check Langau amplitude '''
        mpv = 1.
        x = np.linspace(mpv - 10, mpv + 10, 1000)
        y = pylandau.langau(x, mpv=mpv, eta=1., sigma=1., A=A)
        self.assertAlmostEqual(y.max(), A, delta=1e-4 * A)

    @given(st.floats(constraints.LANDAU_MIN_MPV,
                     constraints.LANDAU_MAX_MPV,
                     allow_nan=False,
                     allow_infinity=False),
           st.floats(constraints.LANDAU_MIN_ETA,
                     constraints.LANDAU_MAX_ETA,
                     allow_nan=False,
                     allow_infinity=False),
           st.floats(constraints.LANDAU_MIN_A,
                     constraints.LANDAU_MAX_A,
                     allow_nan=False,
                     allow_infinity=False))
    def test_landau_stability(self, mpv, eta, A):
        ''' Check Landau outputs for same input parameters '''
        x = np.linspace(mpv - 5 * eta, mpv + 5 * eta, 1000)
        y_1 = pylandau.landau(x, mpv=mpv, eta=eta, A=A)
        y_2 = pylandau.landau(x, mpv=mpv, eta=eta, A=A)
        self.assertTrue(np.all(y_1 == y_2))

    @given(st.floats(constraints.LANGAU_MIN_MPV,
                     constraints.LANGAU_MAX_MPV,
                     allow_nan=False,
                     allow_infinity=False),
           st.floats(constraints.LANGAU_MIN_ETA,
                     constraints.LANGAU_MAX_ETA,
                     allow_nan=False,
                     allow_infinity=False),
           st.floats(constraints.LANGAU_MIN_SIGMA,
                     constraints.LANGAU_MAX_SIGMA,
                     allow_nan=False,
                     allow_infinity=False),
           st.floats(constraints.LANGAU_MIN_A,
                     constraints.LANGAU_MAX_A,
                     allow_nan=False,
                     allow_infinity=False)
           )
    def test_langau_stability(self, mpv, eta, sigma, A):
        ''' Check Langau outputs for same input parameters '''
        # Correct input to avoid oscillations
        if sigma > 100 * eta:
            sigma = eta
        assume(sigma * eta < constraints.LANGAU_MAX_ETA_SIGMA)
        x = np.linspace(mpv - 5 * sigma * eta, mpv + 5 * sigma * eta, 1000)
        y_1 = pylandau.langau(x, mpv=mpv, eta=eta, sigma=sigma, A=A)
        y_2 = pylandau.langau(x, mpv=mpv, eta=eta, sigma=sigma, A=A)
        self.assertTrue(np.all(y_1 == y_2))


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestFuzzing)
    unittest.TextTestRunner(verbosity=2).run(suite)



# Plot the functions using matplotlib

import numpy as np
import pylandau
import matplotlib.pyplot as plt

x = np.arange(0, 100, 0.01)
for A, eta, mpv in ((1, 1, 10), (1, 2, 30), (0.5, 5, 50)):
    # Use the function that calculates y values when given array
    plt.plot(x, pylandau.landau(x, mpv, eta, A), '-', label='mu=%1.1f, eta=%1.1f, A=%1.1f' % (mpv, eta, A))

    # Use the function that calculates the y value given a x value, (e.g. needed for minimizers)
    y = np.array([pylandau.get_landau(x_value, mpv, eta, A) for x_value in x])
    plt.plot(x, y, 'r--', label='mu=%1.1f, eta=%1.1f, A=%1.1f' % (mpv, eta, A))

    # Use the function that calculates the y value given a x value, (e.g. needed for minimizers)
    sigma = 10
    plt.plot(x, pylandau.langau(x, mpv, eta, sigma, A), '--', label='mu=%1.1f, eta=%1.1f, sigma=%1.1f, A=%1.1f' % (A, eta, sigma, mpv))
plt.legend(loc=0)
plt.show()



# Plot Landau distributions with FWHM and MPV
import numpy as np
import matplotlib.pyplot as plt
import pylandau
from scipy.interpolate import splrep, sproot


def fwhm(x, y, k=10):  # http://stackoverflow.com/questions/10582795/finding-the-full-width-half-maximum-of-a-peak
    """
    Determine full-with-half-maximum of a peaked set of points, x and y.

    Assumes that there is only one peak present in the datasset.  The function
    uses a spline interpolation of order k.
    """

    class MultiplePeaks(Exception):
        pass

    class NoPeaksFound(Exception):
        pass

    half_max = np.amax(y) / 2.0
    s = splrep(x, y - half_max)
    roots = sproot(s)

    if len(roots) > 2:
        raise MultiplePeaks("The dataset appears to have multiple peaks, and "
                            "thus the FWHM can't be determined.")
    elif len(roots) < 2:
        raise NoPeaksFound("No proper peaks were found in the data set; likely "
                           "the dataset is flat (e.g. all zeros).")
    else:
        return roots[0], roots[1]

x = np.arange(0, 100, 0.01)
for A, eta, mpv in ((1, 1, 10), (1, 2, 30), (0.5, 5, 50)):
    y = pylandau.landau(x, mpv, eta, A)
    plt.plot(x, y, label='A=%d, mpv=%d, eta=%d' % (A, mpv, eta))
    x_fwhm_1, x_fwhm_2 = fwhm(x, y)
    plt.plot([x_fwhm_1, x_fwhm_2], [np.max(y) / 2., np.max(y) / 2.], label='FWHM: %1.1f' % np.abs(x_fwhm_1 - x_fwhm_2))
    plt.plot([mpv, mpv], [0., np.max(y)], label='MPV: %1.1f' % mpv)
plt.legend(loc=0)
plt.show()



# distutils: language=c++
# distutils: define_macros=CYTHON_TRACE=1
# cython: boundscheck=False
# cython: wraparound=False
# cython: embedsignature=True
# cython: linetrace=True
# cython: language_level=3

import numpy as np
cimport numpy as cnp
cnp.import_array()

from scipy.optimize import fmin


cdef extern from "numpy/arrayobject.h":
    void PyArray_ENABLEFLAGS(cnp.ndarray arr, int flags)


cdef extern from "pylandau_src.cpp":
    double * getLandauPDFData(double * & data, const unsigned int & size, const double & mu, const double & eta) except +
    double * getLangauPDFData(double * & data, const unsigned int & size, const double & mu, const double & eta, const double & sigma) except +
    double landauPDF(const double & x, const double & xi, const double & x0) except +
    double landauGaussPDF(const double & x, const double & mu, const double & eta, const double & sigma) except +
    double gaussPDF(const double & x, const double & mu, const double & sigma)


cdef data_to_numpy_array_double(cnp.double_t * ptr, cnp.npy_intp N):
    cdef cnp.ndarray[cnp.double_t, ndim = 1] arr = cnp.PyArray_SimpleNewFromData(1, < cnp.npy_intp * > & N, cnp.NPY_DOUBLE, < cnp.double_t * > ptr)
    PyArray_ENABLEFLAGS(arr, cnp.NPY_OWNDATA)
    return arr

cdef cnp.double_t * result = NULL


# The pdf are defined by the original algorithm where mu != MPV

def get_landau_pdf(value, mu=0, eta=1):
    return landauPDF(< const double&> value, < const double&> mu, < const double&> eta)


def get_gauss_pdf(value, mu=0, sigma=1):
    return gaussPDF(< const double&> value, < const double&> mu, < const double&> sigma)


def get_langau_pdf(value, mu=0, eta=1, sigma=1):
    return landauGaussPDF( < const double&> value, < const double&> mu, < const double&> eta, < const double&> sigma)


def landau_pdf(cnp.ndarray[cnp.double_t, ndim=1] array, mu=0, eta=1):
    mpv, eta, sigma, _ = _check_parameter(mpv=mu, eta=eta, sigma=0.)
    result = getLandauPDFData( < double*& > array.data, < const unsigned int&> array.shape[0], < const double&> mu, < const double&> eta)
    return data_to_numpy_array_double(result, array.shape[0])


def langau_pdf(cnp.ndarray[cnp.double_t, ndim=1] array, mu=0, eta=1, sigma=1):
    mpv, eta, sigma, _ = _check_parameter(mpv=mu, eta=eta, sigma=sigma)
    result = getLangauPDFData( < double*& > array.data, < const unsigned int&> array.shape[0], < const double&> mu, < const double&> eta, < const double&> sigma)
    if result != NULL:
        return data_to_numpy_array_double(result, array.shape[0])


# These function add a amplitude parameter A and shift the function that mu = MPV
# This is done numerically very simple, stability might suffer

def get_landau(value, mpv=0, eta=1, A=1):
    mpv, eta, sigma, A = _check_parameter(mpv=mpv, eta=eta, sigma=0., A=A)
    mpv_scaled, eta, sigma, A_scaled = _scale_to_mpv(mpv, eta, sigma=0, A=A)

    # Numerical scaling maximum to A
    return get_landau_pdf(value, eta) * A_scaled


def get_langau(value, mpv=0, eta=1, sigma=1, A=1, scale_langau=True):
    mpv, eta, sigma, A = _check_parameter(mpv=mpv, eta=eta, sigma=sigma, A=A)
    mpv_scaled, eta, sigma, A_scaled = _scale_to_mpv(mpv, eta, sigma, A)

    if scale_langau:
        mpv_scaled, _, _, A_scaled = _scale_to_mpv(mpv, eta, sigma, A=A)
    else:
        mpv_scaled, _, _, A_scaled = _scale_to_mpv(mpv, eta, sigma=0, A=A)

    # Numerical scaling maximum to A
    return get_langau_pdf(value, eta, sigma) * A_scaled


def landau(cnp.ndarray[cnp.double_t, ndim=1] array, mpv=0, eta=1, A=1):
    if (A == 0.):
        return np.zeros_like(array)

    mpv, eta, sigma, A = _check_parameter(mpv=mpv, eta=eta, sigma=0., A=A)
    mpv_scaled, eta, sigma, A_scaled = _scale_to_mpv(mpv, eta, sigma=0., A=A)

    # Numerical scaling maximum to A
    return landau_pdf(array, mpv_scaled, eta) * A_scaled


def langau(cnp.ndarray[cnp.double_t, ndim=1] array, mpv=0, eta=1, sigma=1, A=1, scale_langau=True):
    ''' Returns a Landau convoluded with a Gaus.

    If scale_langau is true the Langau function maximum is at mpv with amplitude A.
    Otherwise the Landau function maximum is at mpv with amplitude A, thus not the resulting Langau. '''

    if (A == 0.):
        return np.zeros_like(array)

    mpv, eta, sigma, A = _check_parameter(mpv=mpv, eta=eta, sigma=sigma, A=A)

    if sigma == 0:
        return landau(array, mpv=mpv, eta=eta, A=A)

    if scale_langau:
        mpv_scaled, _, _, A_scaled = _scale_to_mpv(mpv, eta, sigma, A=A)
    else:
        mpv_scaled, _, _, A_scaled = _scale_to_mpv(mpv, eta, sigma=0, A=A)

    # Numerical scaling maximum to A
    return langau_pdf(array, mpv_scaled, eta, sigma) * A_scaled


def _check_parameter(mpv, eta, sigma, A=1.):
    if eta < 1e-9:
        print('WARNING: eta < 1e-9 is not supported. eta set to 1e-9.')
        eta = 1e-9
    if sigma < 0:
        sigma *= -1
    if sigma > 100 * eta:
        print('WARNING: sigma > 100 * eta can lead to oszillations. Check result.')
    if A < 0.:
        raise ValueError('A has to be >= 0')

    return float(mpv), float(eta), float(sigma), float(A)


def _scale_to_mpv(mu, eta, sigma=0., A=None):
    ''' In the original function definition mu != mpv.
    This is fixed here numerically. Also the amplitude is
    scaled to A. '''

    if sigma > 0:
        # https://github.com/SiLab-Bonn/pyLandau/issues/11
        if abs(mu) > 1.:
            x0 = mu
        else:
            x0 = eta * np.sign(mu)
        res = fmin(lambda x: -langau_pdf(x, mu, eta, sigma), x0=x0,
                   full_output=True, disp=False, xtol=0.000001, ftol=0.000001)
    else:
        # https://github.com/SiLab-Bonn/pyLandau/issues/11
        if abs(mu) > 1.:
            x0 = mu
        else:
            x0 = eta * np.sign(mu)
        res = fmin(lambda x: -landau_pdf(x, mu, eta), x0=x0,
                   full_output=True, disp=False, xtol=0.000001, ftol=0.000001)

    if res[4] != 0:
        raise RuntimeError(
            'Cannot calculate MPV, check function parameters and file bug report!')

    if A:
        A = A / -res[1]
        return mu + (mu - res[0]), eta, sigma, A
    else:
        return mu + (mu - res[0]), eta, sigma



''' Constraints that should be followed to ensure stable and proper results '''

# Value input x
MIN_X = - 1e9
MAX_X = 1e9

# pylandau.get_langau_pdf
# pylandau.langau_pdf
LANDAU_PDF_MIN_MU = -1e9
LANDAU_PDF_MAX_MU = 1e9

# pylandau.landau
LANDAU_MIN_MPV = -1e9
LANDAU_MAX_MPV = 1e9
LANDAU_MIN_ETA = 1e-3
LANDAU_MAX_ETA = 1e9
LANDAU_MIN_A = 0.
LANDAU_MAX_A = 1e9

# pylandau.langau
LANGAU_MIN_MPV = -1e9
LANGAU_MAX_MPV = 1e9
LANGAU_MIN_ETA = 1e-1
LANGAU_MAX_ETA = 1e9
LANGAU_MIN_SIGMA = 1e-1
LANGAU_MAX_SIGMA = 1e9
LANGAU_MIN_A = 0.
LANGAU_MAX_A = 1e9
LANGAU_MAX_ETA_SIGMA = 1e9  # sigma * eta



# Plot the Landau Gauss convolution (Langau) and cross check with convolution with scipy

import numpy as np
import pylandau
import matplotlib.pyplot as plt
from scipy.ndimage.filters import gaussian_filter1d

mpv, eta, sigma, A = 10, 1, 3, 1
x = np.arange(0, 50, 0.01)
y = pylandau.landau(x, mpv=mpv, eta=eta, A=A)
y_gconv = gaussian_filter1d(y, sigma=sigma / 0.01)
# Scaled means that the resulting Landau*Gauss function maximum is A at mpv
# Otherwise the underlying Landau function before convolution has the 
# function maximum A at mpv
y_gconv_2 = pylandau.langau(x, mpv, eta, sigma, A, scale_langau=False)
y_gconv_3 = pylandau.langau(x, mpv, eta, sigma, A, scale_langau=True)
plt.plot(x, y, label='Landau')
plt.plot(x, y_gconv_2, label='Langau')
plt.plot(x, y_gconv, '--', label='Langau Scipy')
plt.plot(x, y_gconv_3, label='Langau scaled')
plt.legend(loc=0)
plt.show()



#!/usr/bin/env python
import builtins
from setuptools import setup, find_packages, Extension  # This setup relies on setuptools since distutils is insufficient and badly hacked code
from setuptools.command.build_ext import build_ext as _build_ext


class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        builtins.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

# Check if cython exists, then use it. Otherwise compile already cythonized cpp file
have_cython = False
try:
    from Cython.Build import cythonize
    have_cython = True
except ImportError:
    pass

if have_cython:
    cpp_extension = cythonize(Extension('pylandau', ['pyLandau/cpp/pylandau.pyx']))
else:
    cpp_extension = [Extension('pylandau',
                               sources=['pyLandau/cpp/pylandau.cpp'],
                               language="c++")]

install_requires = ['cython>=0.29', 'numpy>=1.21']

setup(
    cmdclass={'build_ext': build_ext},
    install_requires=install_requires,
    packages=find_packages(),
    include_package_data=True,  # accept all data files and directories matched by MANIFEST.in or found in source control
    package_data={'': ['README.*', 'VERSION'], 'docs': ['*'], 'examples': ['*']},
    ext_modules=cpp_extension,
    keywords=['Landau', 'Langau', 'PDF'],
    platforms='any'
)



''' Advanced fitting example that calculated the MPV errors correctly. 
For this iminuit is used plus an additional step with minos to get
asymmetric correct errors.

Thus iminuit has to be installed (pip install iminuit).

Further info on minuit and the original implementation:
http://seal.web.cern.ch/seal/documents/minuit/mnusersguide.pdf
'''

import numpy as np
import iminuit

from matplotlib import pyplot as plt

from pylandau import langau


def fit_landau_migrad(x, y, p0, limit_mpv, limit_eta, limit_sigma, limit_A):
    def minimizeMe(mpv, eta, sigma, A):
        chi2 = np.sum(np.square(y - langau(x, mpv, eta, sigma, A).astype(float)) / np.square(yerr.astype(float)))
        return chi2 / (x.shape[0] - 5)  # devide by NDF

    # Prefit to get correct errors
    yerr = np.sqrt(y)  # Assume error from measured data
    yerr[y < 1] = 1
    m = iminuit.Minuit(minimizeMe,
                       mpv=p0[0],
                       limit_mpv=limit_mpv,
                       error_mpv=1,
                       eta=p0[1],
                       error_eta=0.1,
                       limit_eta=limit_eta,
                       sigma=p0[2],
                       error_sigma=0.1,
                       limit_sigma=limit_sigma,
                       A=p0[3],
                       error_A=1,
                       limit_A=limit_A,
                       errordef=1,
                       print_level=2)
    m.migrad()

    if not m.get_fmin().is_valid:
        raise RuntimeError('Fit did not converge')

    # Main fit with model errors
    yerr = np.sqrt(langau(x,
                          mpv=m.values['mpv'],
                          eta=m.values['eta'],
                          sigma=m.values['sigma'],
                          A=m.values['A']))  # Assume error from measured data
    yerr[y < 1] = 1

    m = iminuit.Minuit(minimizeMe,
                       mpv=m.values['mpv'],
                       limit_mpv=limit_mpv,
                       error_mpv=1,
                       eta=m.values['eta'],
                       error_eta=0.1,
                       limit_eta=limit_eta,
                       sigma=m.values['sigma'],
                       error_sigma=0.1,
                       limit_sigma=limit_sigma,
                       A=m.values['A'],
                       error_A=1,
                       limit_A=limit_A,
                       errordef=1,
                       print_level=2)
    m.migrad()

    fit_values = m.values

    values = np.array([fit_values['mpv'],
                       fit_values['eta'],
                       fit_values['sigma'],
                       fit_values['A']])

    m.hesse()

    m.minos()
    minos_errors = m.get_merrors()

    if not minos_errors['mpv'].is_valid:
        print('Warning: MPV error determination with Minos failed! You can still use Hesse errors.')

    errors = np.array([(minos_errors['mpv'].lower, minos_errors['mpv'].upper),
                       (minos_errors['eta'].lower, minos_errors['eta'].upper),
                       (minos_errors['sigma'].lower, minos_errors['sigma'].upper),
                       (minos_errors['A'].lower, minos_errors['A'].upper)])

    return values, errors, m

if __name__ == '__main__':
    # Fake counting experiment with Landgaus distribution
    x = np.arange(100).astype(np.float)
    y = langau(x,
               mpv=30.,
               eta=5.,
               sigma=4.,
               A=1000.)
    # Add poisson error
    y += np.random.normal(np.zeros_like(y), np.sqrt(y))

    # Fit the data with numerical exact error estimation
    # Taking into account correlations
    values, errors, m = fit_landau_migrad(x,
                                          y,
                                          p0=[x.shape[0] / 2., 10., 10., np.max(y)],
                                          limit_mpv=(10., 100.),
                                          limit_eta=(2., 20.),
                                          limit_sigma=(2., 20.),
                                          limit_A=(500., 1500.))

    # Plot fit result
    yerr = np.sqrt(langau(x, *values))
    plt.errorbar(x, y, yerr, fmt='.')
    plt.plot(x, langau(x, *values), '-', label='Fit:')
    plt.legend(loc=0)
    plt.show()

    # Show Chi2 as a function of the mpv parameter
    # The cut of the parabola at dy = 1 defines the error
    # https://github.com/iminuit/iminuit/blob/master/tutorial/tutorial.ipynb
    plt.clf()
    m.draw_mnprofile('mpv', subtract_min=False)
    plt.show()

    # Show contour plot, is somewhat correlation / fit confidence
    # of these two parameter. The off regions are a hint for
    # instabilities of the langau function
    plt.clf()
    m.draw_mncontour('mpv', 'sigma', nsigma=3)
    plt.show()



''' Script to check pylandau data type handling.
'''
import unittest

import numpy as np
import pylandau

from hypothesis import given
import hypothesis.extra.numpy as nps
import hypothesis.strategies as st

from tests import constraints


def python_number(draw, min_val, max_val):
    return draw(st.one_of(st.floats(min_val,
                                    max_val,
                                    allow_nan=False,
                                    allow_infinity=False),
                          st.integers(np.ceil(min_val),
                                      np.floor(max_val))))


def numpy_number(draw, min_val, max_val):
    dtype = draw(st.one_of(nps.integer_dtypes(),
                           nps.unsigned_integer_dtypes(),
                           nps.floating_dtypes()))

    if 'f' in dtype.str:
        if min_val < np.finfo(dtype).min:
            min_val = np.finfo(dtype).min
        if max_val > np.finfo(dtype).max:
            max_val = np.finfo(dtype).max
        number = draw(st.floats(min_val, max_val, allow_nan=False,
                                allow_infinity=False))
    else:
        min_val, max_val = np.ceil(min_val), np.floor(max_val)
        if min_val < np.iinfo(dtype).min:
            min_val = np.iinfo(dtype).min
        if max_val > np.iinfo(dtype).max:
            max_val = np.iinfo(dtype).max
        number = draw(st.integers(min_val, max_val))

    return np.array([number], dtype)[0]


@st.composite
def A_landau(draw):
    return draw(st.sampled_from([numpy_number(draw,
                                              constraints.LANDAU_MIN_A,
                                              constraints.LANDAU_MAX_A),
                                 python_number(draw,
                                               constraints.LANDAU_MIN_A,
                                               constraints.LANDAU_MAX_A)]))


@st.composite
def MPV_landau(draw):
    return draw(st.sampled_from([numpy_number(draw,
                                              constraints.LANDAU_MIN_MPV,
                                              constraints.LANDAU_MAX_MPV),
                                 python_number(draw,
                                               constraints.LANDAU_MIN_MPV,
                                               constraints.LANDAU_MAX_MPV)]))


@st.composite
def eta_landau(draw):
    return draw(st.sampled_from([numpy_number(draw,
                                              constraints.LANDAU_MIN_ETA,
                                              constraints.LANDAU_MAX_ETA),
                                 python_number(draw,
                                               constraints.LANDAU_MIN_ETA,
                                               constraints.LANDAU_MAX_ETA)]))


@st.composite
def A_langau(draw):
    return draw(st.sampled_from([numpy_number(draw,
                                              constraints.LANGAU_MIN_A,
                                              constraints.LANGAU_MAX_A),
                                 python_number(draw,
                                               constraints.LANGAU_MIN_A,
                                               constraints.LANGAU_MAX_A)]))


@st.composite
def MPV_langau(draw):
    return draw(st.sampled_from([numpy_number(draw,
                                              constraints.LANGAU_MIN_MPV,
                                              constraints.LANGAU_MAX_MPV),
                                 python_number(draw,
                                               constraints.LANGAU_MIN_MPV,
                                               constraints.LANGAU_MAX_MPV)]))


@st.composite
def eta_langau(draw):
    return draw(st.sampled_from([numpy_number(draw,
                                              constraints.LANGAU_MIN_ETA,
                                              constraints.LANGAU_MAX_ETA),
                                 python_number(draw,
                                               constraints.LANGAU_MIN_ETA,
                                               constraints.LANGAU_MAX_ETA)]))


@st.composite
def sigma_langau(draw):
    return draw(st.sampled_from([numpy_number(draw,
                                              constraints.LANGAU_MIN_SIGMA,
                                              constraints.LANGAU_MAX_SIGMA),
                                 python_number(draw,
                                               constraints.LANGAU_MIN_SIGMA,
                                               constraints.LANGAU_MAX_SIGMA)]))


@st.composite
def x(draw):
    return draw(st.sampled_from([numpy_number(draw,
                                              constraints.MIN_X,
                                              constraints.MAX_X),
                                 python_number(draw,
                                               constraints.MIN_X,
                                               constraints.MAX_X)]))


class TestDatatypes(unittest.TestCase):

    @given(x(), MPV_landau(), eta_landau())
    def test_landau_pdf_inputs(self, x, mpv, eta):
        ''' Check Landau PDF result for different input parameter dtypes'''
        self.assertTrue(isinstance(pylandau.get_landau_pdf(x, mu=mpv,
                                                           eta=eta),
                                   float))

    @given(x(), MPV_landau(), eta_landau(), sigma_langau())
    def test_langau_pdf_inputs(self, x, mpv, eta, sigma):
        ''' Check Langau PDF result for different input parameter dtypes'''
        self.assertTrue(isinstance(pylandau.get_langau_pdf(x, mu=mpv,
                                                           eta=eta,
                                                           sigma=sigma),
                                   float))


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestDatatypes)
    unittest.TextTestRunner(verbosity=2).run(suite)
